/* PulseAudioClip.java
   Copyright (C) 2008 Red Hat, Inc.

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

IcedTea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
 */

package org.classpath.icedtea.pulseaudio;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.Control;
import javax.sound.sampled.Line;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Control.Type;

public class PulseAudioClip implements Clip {

	private byte[] data = null;

	// FIXME
	private int bufferSize = 0;

	// these are frame indices. so counted from 0
	private long currentFrame = 0;
	private int frameCount = 0;
	private long startFrame = 0;
	private long endFrame = 0;
	private long framesSinceOpen = 0;

	private AudioFormat currentFormat = null;
	private static final AudioFormat DEFAULT_FORMAT = new AudioFormat(
			AudioFormat.Encoding.PCM_UNSIGNED, 22050, 8, 2, 2, 22050 / 2, false);

	private boolean isOpen = false;

	private List<LineListener> lineListeners = null;

	private static final int DEFAULT_BUFFER_SIZE = 0;
	public static final String DEFAULT_CLIP_NAME = "Clip";

	private Stream stream;

	private Thread clipLoop = new Thread() {
		@Override
		public void run() {
			while (true) {

			}

		}

	};

	static {
		try {
			String library = new java.io.File(".").getCanonicalPath()
					+ java.io.File.separatorChar
					+ System.mapLibraryName("pulse-java");
			System.out.println(library);
			System.load(library);
		} catch (IOException e) {
			assert ("Loading failed".endsWith("library"));
		}
	}

	public PulseAudioClip() {
		lineListeners = new LinkedList<LineListener>();
	}

	@Override
	public void addLineListener(LineListener listener) {
		lineListeners.add(listener);
	}

	@Override
	public int available() {
		return 0; // a clip always returns 0
	}

	@Override
	public void close() {
		// TODO Auto-generated method stub
		stream.drain();
		stream.disconnect();
		isOpen = false;
	}

	@Override
	public void drain() {
		stream.drain();
	}

	@Override
	public void flush() {
		stream.flush();
	}

	@Override
	public int getBufferSize() {
		return bufferSize;
	}

	@Override
	public Control getControl(Type control) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Control[] getControls() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public AudioFormat getFormat() {
		if (!isOpen) {
			return DEFAULT_FORMAT;
		}
		return currentFormat;
	}

	@Override
	public int getFrameLength() {
		return frameCount;
	}

	@Override
	public int getFramePosition() {
		return (int) framesSinceOpen;
	}

	@Override
	public float getLevel() {
		// TODO Auto-generated method stub
		return AudioSystem.NOT_SPECIFIED;
	}

	@Override
	public javax.sound.sampled.Line.Info getLineInfo() {
		return new Line.Info(this.getClass());
	}

	@Override
	public long getLongFramePosition() {
		return framesSinceOpen;
	}

	@Override
	public long getMicrosecondLength() {
		if (!isOpen) {
			return AudioSystem.NOT_SPECIFIED;
		}
		return frameCount / currentFormat.getFrameSize();
	}

	@Override
	public long getMicrosecondPosition() {
		return framesSinceOpen / currentFormat.getFrameSize();
	}

	@Override
	public boolean isActive() {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public boolean isControlSupported(Type control) {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public boolean isOpen() {
		return isOpen;
	}

	@Override
	public boolean isRunning() {
		// really confused about what this is supposed to do
		return isActive();
	}

	@Override
	public void loop(int count) {
		// TODO Auto-generated method stub

	}

	@Override
	public void open() throws LineUnavailableException {
		throw new IllegalArgumentException("open() on a Clip is not allowed");
	}

	@Override
	public void open(AudioFormat format, byte[] data, int offset, int bufferSize)
			throws LineUnavailableException {

		long contextPointer = EventLoop.getEventLoop().getContextPointer();
		int channels = 2;
		int sampleRate = 22050;
		stream = new Stream(contextPointer, DEFAULT_CLIP_NAME,
				Stream.Format.PA_SAMPLE_U8, sampleRate, channels);

		isOpen = true;
	}

	@Override
	public void open(AudioInputStream stream) throws LineUnavailableException,
			IOException {
		// TODO Auto-generated method stub

		byte[] buffer = new byte[(int) (stream.getFrameLength() * stream
				.getFormat().getFrameSize())];

		open(stream.getFormat(), buffer, 0, DEFAULT_BUFFER_SIZE);

	}

	@Override
	public void removeLineListener(LineListener listener) {
		lineListeners.remove(listener);
	}

	@Override
	public void setFramePosition(int frames) {

		if (frames > frameCount) {
			throw new IllegalArgumentException("incorreft frame value");
		}

		currentFrame = frames;

	}

	@Override
	public void setLoopPoints(int start, int end) {
		if (end == -1) {
			end = frameCount;
		}

		if (end < start) {
			throw new IllegalArgumentException(
					"ending point must be greater than or equal to the starting point");
		}

		startFrame = start;
		endFrame = end;

	}

	@Override
	public void setMicrosecondPosition(long microseconds) {
		float frameIndex = microseconds * currentFormat.getFrameRate();
		currentFrame = (long) frameIndex;

	}

	@Override
	public void start() {
		stream.cork(false);
	}

	@Override
	public void stop() {
		stream.cork(true);
	}

}
