package org.classpath.icedtea.pulseaudio;

import java.util.LinkedList;
import java.util.List;

/**
 * 
 * This class encapsulates a pa_stream object and provides easier access to the
 * native functions
 * 
 */
public class Stream {

	public interface StateListener {
		public void update();
	}

	public interface WriteListener {
		public void update();
	}

	public interface ReadListener {
		public void update();
	}

	public interface OverflowListener {
		public void update();
	}

	public interface UnderflowListener {
		public void update();
	}

	public interface PlaybackStartedListener {
		public void update();
	}

	public interface LatencyUpdateListener {
		public void update();
	}

	public interface MovedListener {
		public void update();
	}

	public interface SuspendedListener {
		public void update();
	}

	public static enum State {
		UNCONNECTED, CREATING, READY, FAILED, TERMINATED,
	}

	public static enum Format {
		PA_SAMPLE_U8, PA_SAMPLE_ULAW, PA_SAMPLE_ALAW, PA_SAMPLE_S16LE, PA_SAMPLE_S16BE, PA_SAMPLE_FLOAT32LE, PA_SAMPLE_FLOAT32BE, PA_SAMPLE_S32LE, PA_SAMPLE_S32BE
	}

	@SuppressWarnings("unused")
	private long streamPointer;

	private Format format;

	private List<StateListener> stateListeners;
	private List<WriteListener> writeListeners;
	private List<ReadListener> readListeners;
	private List<OverflowListener> overflowListeners;
	private List<UnderflowListener> underflowListeners;
	private List<PlaybackStartedListener> playbackStartedListeners;
	private List<LatencyUpdateListener> latencyUpdateListeners;
	private List<MovedListener> movedListeners;
	private List<SuspendedListener> suspendedListeners;

	private native void native_pa_stream_new(long contextPointer, String name,
			String format, int sampleRate, int channels);

	private native int native_pa_stream_get_state();

	private native long native_pa_stream_get_context();

	private native int native_pa_stream_get_index();

	private native int native_pa_stream_get_device_index();

	private native String native_pa_stream_get_device_name();

	private native int native_pa_stream_is_suspended();

	private native int native_pa_stream_connect_playback(String name,
			long buffer_attrPointer, int flags, long volumePointer,
			long sync_streamPointer);

	private native int native_pa_stream_connect_record(String device,
			long buffer_attrPointer, int flags);

	private native int native_pa_stream_disconnect();

	private native int native_pa_stream_write(byte[] data, int offset,
			int length);

	/*
	 * private native int native_pa_stream_peek (pa_stream *p, const void
	 * **data, size_t *nbytes) Read the next fragment from the buffer (for
	 * recording). int pa_stream_drop (pa_stream *p) Remove the current fragment
	 * on record streams.
	 * 
	 */

	private native int native_pa_stream_writable_size();

	private native int native_pa_stream_readable_size();

	private native long native_pa_stream_drain();

	/*
	 * Drain a playback stream. pa_operation pa_stream_update_timing_info
	 * (pa_stream *p, pa_stream_success_cb_t cb, void *userdata) Request a
	 * timing info structure update for a stream.
	 */

	private native long native_pa_stream_cork(int b);

	private native long native_pa_stream_flush();

	/*
	 * pa_operation pa_stream_prebuf (pa_stream *s, pa_stream_success_cb_t cb,
	 * void *userdata) Reenable prebuffering as specified in the pa_buffer_attr
	 * structure.
	 */

	private native long native_pa_stream_trigger();

	/* returns an operationPointer */
	private native long native_pa_stream_set_name(String name);

	/*
	 * Return the current playback/recording time private native int
	 * native_pa_stream_get_time (pa_usec_t r_usec);
	 */

	/*
	 * Return the total stream latency private native int
	 * native_pa_stream_get_latency ( pa_usec_t *r_usec, int *negative);
	 */

	/*
	 * const pa_timing_info * pa_stream_get_timing_info (pa_stream *s) Return
	 * the latest raw timing data structure. const pa_sample_spec *
	 * pa_stream_get_sample_spec (pa_stream *s) Return a pointer to the stream's
	 * sample specification. const pa_channel_map * pa_stream_get_channel_map
	 * (pa_stream *s) Return a pointer to the stream's channel map. const
	 * pa_buffer_attr * pa_stream_get_buffer_attr (pa_stream *s) Return the
	 * per-stream server-side buffer metrics of the stream. pa_operation *
	 * pa_stream_set_buffer_attr (pa_stream *s, const pa_buffer_attr *attr,
	 * pa_stream_success_cb_t cb, void *userdata) Change the buffer metrics of
	 * the stream during playback. pa_operation * pa_stream_update_sample_rate
	 * (pa_stream *s, uint32_t rate, pa_stream_success_cb_t cb, void *userdata)
	 * Change the stream sampling rate during playback. pa_operation *
	 * pa_stream_proplist_update (pa_stream *s, pa_update_mode_t mode,
	 * pa_proplist *p, pa_stream_success_cb_t cb, void *userdata) Update the
	 * property list of the sink input/source output of this stream, adding new
	 * entries. pa_operation * pa_stream_proplist_remove (pa_stream *s, const
	 * char *const keys[], pa_stream_success_cb_t cb, void *userdata) Update the
	 * property list of the sink input/source output of this stream, remove
	 * entries. int pa_stream_set_monitor_stream (pa_stream *s, uint32_t
	 * sink_input_idx) For record streams connected to a monitor source: monitor
	 * only a very specific sink input of the sink. uint32_t
	 * pa_stream_get_monitor_stream (pa_stream *s) Return what has been set with
	 * pa_stream_set_monitor_stream() ebfore.
	 */

	public Stream(long contextPointer, String name, Format format,
			int sampleRate, int channels) {
		System.out.println("format: " + format.toString());

		stateListeners = new LinkedList<StateListener>();
		writeListeners = new LinkedList<WriteListener>();
		readListeners = new LinkedList<ReadListener>();
		overflowListeners = new LinkedList<OverflowListener>();
		underflowListeners = new LinkedList<UnderflowListener>();
		playbackStartedListeners = new LinkedList<PlaybackStartedListener>();
		latencyUpdateListeners = new LinkedList<LatencyUpdateListener>();
		movedListeners = new LinkedList<MovedListener>();
		suspendedListeners = new LinkedList<SuspendedListener>();

		this.format = format;

		native_pa_stream_new(contextPointer, name, format.toString(),
				sampleRate, channels);
	}

	public void addStateListener(StateListener listener) {
		synchronized (stateListeners) {
			stateListeners.add(listener);
		}
	}

	public void removeStateListener(StateListener listener) {
		synchronized (stateListeners) {
			stateListeners.remove(listener);
		}

	}

	public void addWriteListener(WriteListener listener) {
		synchronized (writeListeners) {
			writeListeners.add(listener);
		}
	}

	public void removeWriteListener(WriteListener listener) {
		synchronized (writeListeners) {
			writeListeners.remove(listener);
		}
	}

	public void addReadListener(ReadListener listener) {
		synchronized (readListeners) {
			readListeners.add(listener);
		}
	}

	public void removeReadListener(ReadListener listener) {
		synchronized (readListeners) {
			readListeners.remove(listener);
		}
	}

	public void addOverflowListener(OverflowListener listener) {
		overflowListeners.add(listener);
	}

	public void removeOverflowListener(OverflowListener listener) {
		overflowListeners.remove(listener);
	}

	public void addUnderflowListener(UnderflowListener listener) {
		underflowListeners.add(listener);
	}

	public void removeUnderflowListener(UnderflowListener listener) {
		underflowListeners.remove(listener);
	}

	public void addPlaybackStartedListener(PlaybackStartedListener listener) {
		playbackStartedListeners.add(listener);
	}

	public void removePlaybackStartedListener(PlaybackStartedListener listener) {
		playbackStartedListeners.remove(listener);
	}

	public void addLatencyUpdateListener(LatencyUpdateListener listener) {
		latencyUpdateListeners.add(listener);
	}

	public void removeLatencyUpdateListener(LatencyUpdateListener listener) {
		latencyUpdateListeners.remove(listener);
	}

	public void addMovedListener(MovedListener listener) {
		movedListeners.add(listener);
	}

	public void removeMovedListener(MovedListener listener) {
		movedListeners.remove(listener);
	}

	public void addSuspendedListener(SuspendedListener listener) {
		suspendedListeners.add(listener);
	}

	public void removeSuspendedListener(SuspendedListener listener) {
		suspendedListeners.remove(listener);
	}

	public Stream.State getState() {
		int state = native_pa_stream_get_state();
		switch (state) {
		case 0:
			return State.UNCONNECTED;
		case 1:
			return State.CREATING;
		case 2:
			return State.READY;
		case 3:
			return State.FAILED;
		case 4:
			return State.TERMINATED;
		default:
			throw new IllegalStateException("invalid stream state");
		}

	}

	public long getContextPointer() {
		return native_pa_stream_get_context();
	}

	public int getSinkInputIndex() {
		return native_pa_stream_get_index();
	}

	/**
	 * 
	 * @return the index of the sink or source this stream is connected to in
	 *         the server
	 */
	public int getDeviceIndex() {
		return native_pa_stream_get_device_index();
	}

	/**
	 * 
	 * @return the name of the sink or source this stream is connected to in the
	 *         server
	 */
	public String getDeviceName() {
		return native_pa_stream_get_device_name();
	}

	/**
	 * if the sink or source this stream is connected to has been suspended.
	 * 
	 * @return
	 */
	public boolean isSuspended() {
		return (native_pa_stream_is_suspended() != 0);
	}

	/**
	 * Connect the stream to a sink
	 * 
	 * @param deviceName
	 */
	public void connectForPlayback(String deviceName) {

		int returnValue = native_pa_stream_connect_playback(deviceName, 0, 0,
				0, 0);
		assert(returnValue == 0);
	}

	/**
	 * Connect the stream to a source.
	 * 
	 */
	public void connectForRecording(String deviceName) {
		int returnValue = native_pa_stream_connect_record(deviceName, 0, 0);
		assert(returnValue == 0);
	}

	/**
	 * Disconnect a stream from a source/sink.
	 */
	public void disconnect() {
		int returnValue = native_pa_stream_disconnect();
		assert(returnValue == 0);
	}	

	/**
	 * Write data to the server
	 * 
	 * @param data
	 * @param length
	 * @return
	 */
	public int write(byte[] data, int offset, int length) {
		return native_pa_stream_write(data, offset, length);
	}

	/**
	 * Read the next fragment from the buffer (for recording).
	 * 
	 * 
	 * @param data
	 */
	public void peek(byte[] data) {

	}

	/**
	 * 
	 * Remove the current fragment on record streams.
	 */
	public void drop() {

	}

	/**
	 * Return the number of bytes that may be written using write().
	 * 
	 * @return
	 */
	public int getWritableSize() {
		return native_pa_stream_writable_size();
	}

	/**
	 * Return the number of bytes that may be read using peek().
	 * 
	 * @return
	 */
	public int getReableSize() {
		return native_pa_stream_readable_size();
	}

	/**
	 * Drain a playback stream
	 * 
	 * @return
	 */
	public Operation drain() {
		Operation drainOperation = new Operation(native_pa_stream_drain());
		return drainOperation;
	}

	/**
	 * this function is called whenever the state changes
	 */
	@SuppressWarnings("unused")
	private void stateCallback() {
		synchronized (stateListeners) {
			for (StateListener listener : stateListeners) {
				listener.update();
			}
		}
	}

	@SuppressWarnings("unused")
	private void writeCallback() {
		synchronized (writeListeners) {
			for (WriteListener listener : writeListeners) {
				listener.update();
			}
		}
	}

	@SuppressWarnings("unused")
	private void readCallback() {
		synchronized (readListeners) {
			for (ReadListener listener : readListeners) {
				listener.update();
			}
		}
	}

	@SuppressWarnings("unused")
	private void overflowCallback() {
		synchronized (overflowListeners) {
			for (OverflowListener listener : overflowListeners) {
				listener.update();
			}
		}
	}

	@SuppressWarnings("unused")
	private void underflowCallback() {
		synchronized (underflowListeners) {
			for (UnderflowListener listener : underflowListeners) {
				listener.update();
			}
		}
	}

	/**
	 * callback function that is called when a the server starts playback after
	 * an underrun or on initial startup
	 */
	@SuppressWarnings("unused")
	private void playbackStartedCallback() {
		assert(false);
		synchronized (playbackStartedListeners) {
			for (PlaybackStartedListener listener : playbackStartedListeners) {
				listener.update();
			}
		}
	}

	/**
	 * called whenever a latency information update happens
	 */
	@SuppressWarnings("unused")
	private void latencyUpdateCallback() {
		synchronized (latencyUpdateListeners) {
			for (LatencyUpdateListener listener : latencyUpdateListeners) {
				listener.update();
			}
		}
	}

	/**
	 * whenever the stream is moved to a different sink/source
	 */
	@SuppressWarnings("unused")
	private void movedCallback() {
		synchronized (movedListeners) {
			for (MovedListener listener : movedListeners) {
				listener.update();
			}
		}
	}

	/**
	 * whenever the sink/source this stream is connected to is suspended or
	 * resumed
	 */
	@SuppressWarnings("unused")
	private void suspendedCallback() {
		synchronized (suspendedListeners) {
			for (SuspendedListener listener : suspendedListeners) {
				listener.update();
			}
		}
	}

	/**
	 * Pause (or resume) playback of this stream temporarily.
	 * 
	 * @param cork
	 * @return
	 */
	public Operation cork(boolean cork) {
		int yes = cork ? 1 : 0;
		Operation corkOperation = new Operation(native_pa_stream_cork(yes));
		return corkOperation;
	}

	public Operation cork() {
		return cork(true);
	}

	public Operation unCork() {
		return cork(false);
	}

	/**
	 * Flush the playback buffer of this stream.
	 * 
	 * @return
	 */
	public Operation flush() {
		Operation flushOperation = new Operation(native_pa_stream_flush());
		return flushOperation;
	}

	/*
	 * Operation pa_stream_prebuf (pa_stream *s, pa_stream_success_cb_t cb, void
	 * *userdata)
	 * 
	 * Reenable prebuffering as specified in the pa_buffer_attr structure.
	 */

	/**
	 * Request immediate start of playback on this stream.
	 */
	public Operation triggerStart() {
		Operation triggerOperation = new Operation(native_pa_stream_trigger());
		return triggerOperation;
	}

	/**
	 * set the stream's name
	 * 
	 * @param name
	 * @return
	 */
	public Operation setName(String name) {
		Operation setNameOperation = new Operation(
				native_pa_stream_set_name(name));
		return setNameOperation;
	}

	public long getTimeInMicroseconds() {
		return -1;
	}

	// TODO: huh?
	/**
	 * @returns the total stream latency
	 */
	public int getLatency() {
		return -1;
	}

	/*
	 * const pa_timing_info * pa_stream_get_timing_info (pa_stream *s) Return
	 * the latest raw timing data structure.
	 * 
	 */

	public Format getFormat() {
		return format;
	}

	/*
	 * const pa_channel_map * pa_stream_get_channel_map (pa_stream *s) Return a
	 * pointer to the stream's channel map.
	 */

	/*
	 * const pa_buffer_attr * pa_stream_get_buffer_attr (pa_stream *s) Return
	 * the per-stream server-side buffer metrics of the stream.
	 * 
	 */

	/*
	 * pa_operation * pa_stream_set_buffer_attr (pa_stream *s, const
	 * pa_buffer_attr *attr, pa_stream_success_cb_t cb, void *userdata) Change
	 * the buffer metrics of the stream during playback.
	 * 
	 */

	/**
	 * Change the stream sampling rate during playback.
	 * 
	 */
	/*
	 * Operation updateSampleRate(int rate) { Operation updateOp = new
	 * Operation(native_pa_stream_) }
	 */

	public long getStreamPointer() {
		return streamPointer;
	}

}
