#include "org_classpath_icedtea_pulseaudio_Stream.h"

#include "jni-common.h"
#include <pulse/pulseaudio.h>
#include <string.h>

typedef struct java_context {
	JNIEnv* env;
	jobject obj;
} java_context;

extern JNIEnv* pulse_thread_env;

static void stream_state_callback(pa_stream* stream, void *userdata) {
	printf("stream_state_callback called\n");

	java_context* context = userdata;
	assert(stream);
	assert(context);
	assert(context->env);
	assert(context->obj);

	if (pa_stream_get_state(stream) == PA_STREAM_CREATING) {
		callJavaVoidMethod(context->env, context->obj, "stateCallback");
	} else {
		callJavaVoidMethod(pulse_thread_env, context->obj, "stateCallback");
	}

}

static void stream_write_callback(pa_stream *stream, size_t length,
		void *userdata) {
	//	printf("stream_write_callback called\n");

	java_context* context = userdata;
	assert(stream);
	assert(context);
	assert(context->env);
	assert(context->obj);

	if (pa_stream_get_state(stream) == PA_STREAM_CREATING) {
		callJavaVoidMethod(context->env, context->obj, "writeCallback");
	} else {
		callJavaVoidMethod(pulse_thread_env, context->obj, "writeCallback");
	}
}

static void stream_read_callback(pa_stream *stream, size_t length,
		void *userdata) {
	printf("stream_read_callback called\n");

	java_context* context = userdata;
	assert(stream);
	assert(context);
	assert(context->env);
	assert(context->obj);

	if (pa_stream_get_state(stream) == PA_STREAM_CREATING) {
		callJavaVoidMethod(context->env, context->obj, "readCallback");
	} else {
		callJavaVoidMethod(pulse_thread_env, context->obj, "readCallback");
	}

}

static void stream_overflow_callback(pa_stream *stream, void *userdata) {
	printf("stream_overflow_callback called\n");

	java_context* context = userdata;
	assert(stream);
	assert(context);
	assert(context->env);
	assert(context->obj);

	if (pa_stream_get_state(stream) == PA_STREAM_CREATING) {
		callJavaVoidMethod(context->env, context->obj, "overflowCallback");
	} else {
		callJavaVoidMethod(pulse_thread_env, context->obj, "overflowCallback");
	}
}

static void stream_underflow_callback(pa_stream *stream, void *userdata) {
	printf("stream_underflow_callback called\n");

	java_context* context = userdata;
	assert(stream);
	assert(context);
	assert(context->env);
	assert(context->obj);

	if (pa_stream_get_state(stream) == PA_STREAM_CREATING) {
		callJavaVoidMethod(context->env, context->obj, "underflowCallback");
	} else {
		callJavaVoidMethod(pulse_thread_env, context->obj, "underflowCallback");
	}
}

// requires pulseaudio 0.9.11 :(
//static void stream_started_callback(pa_stream *p, void *userdata) {
//	printf("stream_started_callback called\n");
//}

static void stream_latency_update_callback(pa_stream *stream, void *userdata) {
	printf("stream_latency_update_callback called\n");

	java_context* context = userdata;
	assert(stream);
	assert(context);
	assert(context->env);
	assert(context->obj);

	if (pa_stream_get_state(stream) == PA_STREAM_CREATING) {
		callJavaVoidMethod(context->env, context->obj, "latencyUpdateCallback");
	} else {
		callJavaVoidMethod(pulse_thread_env, context->obj,
				"latencyUpdateCallback");
	}
}

static void stream_moved_callback(pa_stream *stream, void *userdata) {
	printf("stream_moved_callback called\n");

	java_context* context = userdata;
	assert(stream);
	assert(context);
	assert(context->env);
	assert(context->obj);

	if (pa_stream_get_state(stream) == PA_STREAM_CREATING) {
		callJavaVoidMethod(context->env, context->obj, "movedCallback");
	} else {
		callJavaVoidMethod(pulse_thread_env, context->obj, "movedCallback");
	}

}

static void stream_suspended_callback(pa_stream *stream, void *userdata) {
	printf("stream_suspended_callback called\n");

	java_context* context = userdata;
	assert(stream);
	assert(context);
	assert(context->env);
	assert(context->obj);

	if (pa_stream_get_state(stream) == PA_STREAM_CREATING) {
		callJavaVoidMethod(context->env, context->obj, "suspendedCallback");
	} else {
		callJavaVoidMethod(pulse_thread_env, context->obj, "suspendedCallback");
	}

}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_new
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1new
(JNIEnv* env, jobject obj, jlong contextPointer, jstring nameString, jstring encodingString, jint sampleRate, jint channels) {

	java_context* j_context = malloc(sizeof(java_context));
	j_context->env = env;
	j_context->obj = (*env)->NewGlobalRef(env, obj);

	pa_context* context = convertJavaLongToPointer(contextPointer);
	assert(context);
	const char* name = NULL;
	if (nameString) {
		name = (*env)->GetStringUTFChars(env,nameString, NULL);
		if (name == NULL) {
			return; // oome thrown
		}
	}

	const char *encoding = (*env)->GetStringUTFChars(env, encodingString, NULL);
	if( encoding == NULL) {
		return; //oome thrown
	}

	pa_sample_spec sample_spec;

	if (strcmp(encoding, "PA_SAMPLE_U8") == 0) {
		sample_spec.format = PA_SAMPLE_U8;
	} else if (strcmp(encoding, "PA_SAMPLE_ALAW") == 0) {
		sample_spec.format = PA_SAMPLE_ALAW;
	} else if (strcmp(encoding, "PA_SAMPLE_ULAW;") == 0) {
		sample_spec.format = PA_SAMPLE_ULAW;
	} else if (strcmp(encoding, "PA_SAMPLE_S16BE") == 0) {
		sample_spec.format = PA_SAMPLE_S16BE;
	} else if (strcmp(encoding, "PA_SAMPLE_S16LE") == 0) {
		sample_spec.format = PA_SAMPLE_S16LE;
	} else if (strcmp(encoding, "PA_SAMPLE_S32BE") == 0) {
		sample_spec.format = PA_SAMPLE_S32BE;
	} else if (strcmp(encoding, "PA_SAMPLE_S32LE") == 0) {
		sample_spec.format = PA_SAMPLE_S32LE;
	} else {
		printf("error in open: encoding is : %s\n", encoding);
		throwByName(env, "java/lang/IllegalArgumentException", "Invalid format");
		/* clean up */
		(*env)->ReleaseStringUTFChars(env, encodingString, encoding);
		return;
	}

	sample_spec.rate = sampleRate;
	sample_spec.channels = channels;

	if ( !pa_sample_spec_valid(&sample_spec)) {
		throwByName(env, "java/lang/IllegalArgumentException", "Invalid format");
		(*env)->ReleaseStringUTFChars(env, encodingString, encoding);
		if (name) {
			(*env)->ReleaseStringUTFChars(env, nameString,name);
		}
		return;
	}

	pa_stream* stream = pa_stream_new(context, name, &sample_spec, NULL);
	assert(stream);
	if (name) {
		(*env)->ReleaseStringUTFChars(env, nameString,name);
	}

	setJavaPointer(env, obj, "streamPointer", stream);

	/*
	 * 
	 * The stream has been created; now setup the callbacks 
	 * so we can do somethig about them
	 * 
	 */

	pa_stream_set_state_callback (stream, stream_state_callback, j_context);
	pa_stream_set_write_callback (stream, stream_write_callback, j_context);
	pa_stream_set_read_callback (stream, stream_read_callback, j_context);
	pa_stream_set_overflow_callback (stream, stream_overflow_callback, j_context);
	pa_stream_set_underflow_callback (stream, stream_underflow_callback, j_context);
	//	pa_stream_set_started_callback (stream, stream_started_callback, j_context);
	pa_stream_set_latency_update_callback (stream, stream_latency_update_callback, j_context);
	pa_stream_set_moved_callback (stream, stream_moved_callback, j_context);
	pa_stream_set_suspended_callback (stream, stream_suspended_callback, j_context);

}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_get_state
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1get_1state
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*) getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	return pa_stream_get_state(stream);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_get_context
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1get_1context
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*) getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	pa_context* context = pa_stream_get_context(stream);
	assert(context);
	return convertPointerToJavaLong(context);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_get_index
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1get_1index
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*) getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	return pa_stream_get_index(stream);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_get_device_index
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1get_1device_1index
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*) getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	return pa_stream_get_device_index(stream);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_get_device_name
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1get_1device_1name
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	const char* name = pa_stream_get_device_name(stream);
	assert(name);
	return (*env)->NewStringUTF(env, name);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_is_suspended
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1is_1suspended
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*) getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	return pa_stream_is_suspended(stream);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_connect_playback
 * Signature: (Ljava/lang/String;JIJJ)I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1connect_1playback
(JNIEnv* env, jobject obj, jstring device, jlong pointer, jint anotherPointer, jlong yetAnotherPointer, jlong TooManyPointers) {
	pa_stream* stream = (pa_stream*) getJavaPointer(env, obj, "streamPointer");
	const char* dev = NULL;
	if (device != NULL) {
		dev = (*env)->GetStringUTFChars(env, device, NULL);
		if (dev == NULL) {
			return -1; // oome thrown
		}
	}
	int value = pa_stream_connect_playback(stream, dev, NULL, 0, NULL, NULL);
	if (dev != NULL) {
		(*env)->ReleaseStringUTFChars(env, device, dev);
		dev = NULL;
	}
	return value;
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_connect_record
 * Signature: (Ljava/lang/String;JI)I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1connect_1record
(JNIEnv* env, jobject obj, jstring device, jlong bufferPointer, jint flags) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	const char* dev = NULL;
	if (device != NULL) {
		dev = (*env)->GetStringUTFChars(env, device, NULL);
		if (dev == NULL) {
			return -1; // oome thrown
		}
	}
	pa_buffer_attr* buffer = convertJavaLongToPointer(bufferPointer);
	assert(buffer == NULL);
	int value = pa_stream_connect_record(stream, dev, buffer, flags);
	if (dev != NULL) {
		(*env)->ReleaseStringUTFChars(env, device, dev);
		dev = NULL;
	}
	return value;

}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_disconnect
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1disconnect
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*) getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	return pa_stream_disconnect(stream);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_write
 * Signature: ([BI)I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1write
(JNIEnv* env, jobject obj, jbyteArray data, jint offset, jint data_length) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	assert(stream);
	jbyte* data_buffer = (*env)->GetByteArrayElements(env, data, NULL);
	if (data_buffer == NULL) {
		return -1; // oome thrown
	}
	jbyte* buffer_start = data_buffer + offset;
	int value = pa_stream_write(stream, buffer_start, data_length, NULL, 0, PA_SEEK_RELATIVE);
	(*env)->ReleaseByteArrayElements(env, data, data_buffer, 0);
	return value;
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_writable_size
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1writable_1size
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	size_t size = pa_stream_writable_size(stream);
	return size;

}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_readable_size
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1readable_1size
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	return pa_stream_readable_size(stream);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_drain
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1drain
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	pa_operation* operation = pa_stream_drain(stream, NULL, NULL);
	return convertPointerToJavaLong(operation);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_cork
 * Signature: (I)J
 */
JNIEXPORT jlong JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1cork
(JNIEnv* env, jobject obj, jint yes) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	pa_operation* operation = pa_stream_cork(stream, yes, NULL, NULL);
	return convertPointerToJavaLong(operation);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_flush
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1flush
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*) getJavaPointer(env, obj, "streamPointer");
	pa_operation* operation = pa_stream_flush(stream, NULL, NULL);
	return convertPointerToJavaLong(operation);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_trigger
 * Signature: ()J
 */
JNIEXPORT jlong JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1trigger
(JNIEnv* env, jobject obj) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	pa_operation* operation = pa_stream_trigger(stream, NULL, NULL);
	return convertPointerToJavaLong(operation);
}

/*
 * Class:     org_classpath_icedtea_pulseaudio_Stream
 * Method:    native_pa_stream_set_name
 * Signature: (Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_org_classpath_icedtea_pulseaudio_Stream_native_1pa_1stream_1set_1name
(JNIEnv* env, jobject obj, jstring newName) {
	pa_stream* stream = (pa_stream*)getJavaPointer(env, obj, "streamPointer");
	const char* name;
	name = (*env)->GetStringUTFChars(env, newName, NULL);
	if (name == NULL) {
		return 0; // OutOfMemoryError already thrown
	}

	pa_operation* operation = pa_stream_set_name(stream, name, NULL, NULL);
	assert(operation);
	(*env)->ReleaseStringUTFChars(env, newName, name);

	return convertPointerToJavaLong(operation);
}

