/* PulseAudioClipTest.java
   Copyright (C) 2008 Red Hat, Inc.

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

IcedTea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
 */

package org.classpath.icedtea.pulseaudio;

import java.io.File;
import java.io.IOException;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.Line;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.sound.sampled.DataLine.Info;

import junit.framework.JUnit4TestAdapter;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class PulseAudioClipTest {

	Mixer mixer;

	public static junit.framework.Test suite() {
		return new JUnit4TestAdapter(PulseAudioClipTest.class);
	}

	@Before
	public void setUp() throws LineUnavailableException {
		Mixer.Info wantedMixerInfo = null;
		Mixer.Info[] mixerInfos = AudioSystem.getMixerInfo();
		for (Mixer.Info mixerInfo : mixerInfos) {
			if (mixerInfo.getName().contains("PulseAudio")) {
				wantedMixerInfo = mixerInfo;
				break;
			}
		}
		assert (wantedMixerInfo != null);
		mixer = AudioSystem.getMixer(wantedMixerInfo);
		mixer.open();
	}

	@Test
	public void testObtainingAClip() throws LineUnavailableException {
		Clip clip = (Clip) mixer.getLine(new Line.Info(Clip.class));
		Assert.assertNotNull(clip);
	}

	
	@Test (expected=IllegalArgumentException.class)
	public void testClipOpen() throws LineUnavailableException {
		Clip clip = (Clip) mixer.getLine(new Line.Info(Clip.class));
		clip.open();
	}
	
	@Test
	public void testClipOpens() throws LineUnavailableException, UnsupportedAudioFileException, IOException {
		Clip clip = (Clip) mixer.getLine(new Line.Info(Clip.class));
		File soundFile = new File("testsounds/startup.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		clip.open(audioInputStream);
	}
	
	@Test
	public void testPlayClip() throws LineUnavailableException, IOException, UnsupportedAudioFileException {
		Clip clip = (Clip) mixer.getLine(new Line.Info(Clip.class));
		File soundFile = new File("testsounds/startup.wav");
		AudioInputStream audioInputStream = AudioSystem
				.getAudioInputStream(soundFile);
		clip.open(audioInputStream);
		clip.loop(1);
		
		
		
	}
	
	@Test (expected=IllegalArgumentException.class)
	public void testOpenWrongUse() throws LineUnavailableException {
		Clip clip = (Clip) mixer.getLine(new Line.Info(Clip.class));
		clip.open();
		
	}
	
	@After
	public void tearDown() {

	}

}
